import es, urllib, path, base64, time

info = es.AddonInfo()
info.version = '1.1.0'
info.author  = 'Ojii'
es.ServerVar('globalconfigs_version',info.version,'Version of Globalconfigs').makepublic()

gamedir = path.path(str(es.ServerVar('eventscripts_gamedir')))
owndir  = path.path(str(es.getAddonPath('globalconfigs')))

init = False

def load():
    es.regcmd('gc_sync','globalconfigs/cmd_sync','Synchronise your configs')
    es.regcmd('gc_add','globalconfigs/cmd_add','Add a file to the syncrhonisation list')
    es.regcmd('gc_del','globalconfigs/cmd_del','Delete a file from the synchronisation list')
    es.regcmd('gc_view','globalconfigs/cmd_view','View the files globalconfigs manages')
    es.dbgmsg(0,'Please configure your GlobalConfigs addon before using it!')
    
def es_map_start(ev):
    dcfg.sync()
    initializse()
    if dcfg['mapstart'] == 'send':
        sync_send
    elif dcfg['mapstart'] == 'get':
        sync_get()
    
def initialize():
    global init, own, local, web, sqlite3
    init = True
    local = False
    web = False
    sqlite = False
    if dcfg.as_bool('sqlite'):
        import sqlite3
        sqlite = True
        own = Sqlite(gamedir)
    else:
        own = FileDB(gamedir)
    if dcfg.as_bool('local'):
        if sqlite:
            local = Sqlite(dcfg.as_path('localpath'))
        else:
            local = FileDB(dcfg.as_path('localpath'))
    else:
        web = Web()
        
def cmd_add():
    if not init:
        initialize()
    if not es.getargc() == 2:
        es.dbgmsg(0,'Invalid Syntax! Use: gc_add "path/to/file"')
        return
    filepath = es.getargv(1).replace('\\','/')
    if filepath[0] == '/':
        filepath = filepath[1:]
    own.add(filepath)
    
def cmd_del():
    if not init:
        initialize()
    if not es.getargc() == 2:
        es.dbgmsg(0,'Invalid Syntax! Use: gc_del "path/to/file"')
        return
    filepath = es.getargv(1).replace('\\','/')
    if filepath[0] == '/':
        filepath = filepath[1:]
    own.delete(filepath)
    
def cmd_view():
    if not init:
        initialize()
    filelist = own.getFilelist()
    es.dbgmsg(0,'GlobalConfigs filelist:')
    es.dbgmsg(0,'-----------------------')
    for f in filelist:
        es.dbgmsg(0,f.split(' ')[1])
    
def cmd_sync():
    if not init:
        initialize()
    if es.getargc() == 2 and es.getargv(1) == 'send':
        sync_send()
    else:
        sync_get()
    es.dbgmsg(0,'Synchronisation done')
        
def sync_send():
    for filepath in own.getChanged():
        if local:
            local.send(filepath)
        else:
            web.send(filepath)
        es.dbgmsg(0,'Sent file %s' % filepath)

def sync_get():
    if web: # Check news from the web
        filelist = web.getFilelist()
        for file in filelist:
            if not own.check(file):
                web.getFile(file)
                es.dbgmsg(0,'Getting file from web: %s' % file)
    else: # Check news from local
        filelist = local.getFilelist()
        for file in filelist:
            if not own.check(file):
                local.getFile(file)
                es.dbgmsg(0,'Getting file local: %s' % file)
    for file in own.getFilelist(): # Send own unique files
        print file
        print filelist
        if not file in filelist:
            if local:
                local.send(file.split(' ',1)[1])
            if web:
                web.send(file.split(' ',1)[1])
            es.dbgmsg(0,'Sent file %s' % file)
            
def makeName(filepath):
    return filepath.replace('/','_').replace('.','_')

### WEB INTERFACE ###    
class Web(object):
    def __init__(self):
        self.url = dcfg['url']
        self.pw = dcfg['password']
        # somehow should send servername + ip:port...
        
    def send(self,filepath):
        content = base64.b64encode(gamedir.joinpath(filepath).text())
        params = urllib.urlencode([
                ('version',info.version),
                ('pw',self.pw),
                ('servername',str(es.ServerVar('hostname'))),
                ('port',str(es.ServerVar('hostport'))),
                ('action','send'),
                ('filepath',filepath),
                ('b64',content)
        ])
        u = urllib.urlopen(self.url,params)
        u.read()
        
    def getFilelist(self):
        params = urllib.urlencode([
                ('version',info.version),
                ('pw',self.pw),
                ('servername',str(es.ServerVar('hostname'))),
                ('port',str(es.ServerVar('hostport'))),
                ('action','getfilelist')
        ])
        u = urllib.urlopen(self.url,params)
        return u.read().split('\\n')
    
    def getFile(self,webanswer):
        timestamp,filepath = webanswer.split(' ',1)
        timestamp = float(timestamp)
        name = makeName(filepath)
        params = [
                ('version',info.version),
                ('pw',self.pw),
                ('servername',str(es.ServerVar('hostname'))),
                ('port',str(es.ServerVar('hostport'))),
                ('action','getfile'),
                ('filepath',filepath)
        ]
        for p in params:
            print p
        params = urllib.urlencode(params)
        u = urllib.urlopen(self.url,params)
        response = u.read()
        print response
        content = base64.b64decode(response)
        f = gamedir.joinpath(filepath)
        f.write_text(content)
        own.update(name,timestamp,filepath)
        
### SQLITE INTERFACE ###
class Sqlite(object):
    def __init__(self,dbpath):
        print dbpath
        self.con = sqlite3.connect(dbpath.joinpath('addons/eventscripts/globalconfigs/data/database.db'))
        self.gamedir = dbpath
        self.cur = self.con.cursor()
        self.cur.execute("CREATE TABLE IF NOT EXISTS globalconfigs (name TEXT PRIMARY KEY, timestamp INTEGER, filepath TEXT)")
        self.con.commit()
        
    def __contains__(self,name):
        self.cur.execute("SELECT filepath FROM globalconfigs WHERE name='%s'" % name)
        result = self.cur.fetchone()
        return bool(result)
        
    def add(self,filepath):
        filepath = filepath
        name = makeName(filepath)
        timestamp = 0
        self.cur.execute("INSERT INTO globalconfigs (name,timestamp,filepath) VALUES ('%s',%s,'%s')" % (name,timestamp,filepath))
        self.con.commit()
        
    def delete(self,filepath):
        filepath = filepath
        name = makeName(filepath)
        self.cur.execute("DELETE FROM globalconfigs WHERE name='%s'" % name)
        
    def getChanged(self):
        changed = []
        self.cur.execute("SELECT timestamp,filepath FROM globalconfigs")
        result = self.cur.fetchall()
        for file in result:
            timestamp,filepath = file
            fpath = self.gamedir.joinpath(filepath)
            if int(fpath.getmtime()) != timestamp:
                changed.append(filepath)
        return changed 
    
    def check(self,webanswer):
        timestamp,filepath = webanswer.split(' ',1)
        name = makeName(filepath)
        if not name in self:
            return False
        timestamp = int(timestamp)
        self.cur.execute("SELECT timestamp FROM globalconfigs WHERE name='%s'" % name)
        result = self.cur.fetchone()
        return result[0] == timestamp # False= file will be updated!
    
    def getFilelist(self):
        self.iUpdate()
        self.cur.execute("SELECT timestamp,filepath FROM globalconfigs")
        result = self.cur.fetchall()
        return map(lambda x: '%s %s' % x,result)
    
    def iUpdate(self):
        changed = self.getChanged()
        for filepath in changed:
            name = makeName(filepath)
            timestamp = int(self.gamedir.joinpath(filepath).getmtime())
            self.update(name,timestamp,filepath)
    
    def update(self,name,timestamp,filepath):
        if name in self:
            self.cur.execute("UPDATE globalconfigs SET timestamp=%s WHERE name='%s'" % (timestamp,name))
        else:
            self.cur.execute("INSERT INTO globalconfigs (name,timestamp,filepath) VALUES ('%s',%s,'%s')" % (name,timestamp,filepath))
        self.con.commit()
        
    def getFile(self,webanswer):
        timestamp,filepath = webanswer.split(' ',1)
        timestamp = int(timestamp)
        name = makeName(filepath)
        src = self.gamedir.joinpath(filepath)
        dst = gamedir.joinpath(filepath)
        dst.write_text(src.text())
        own.update(name,timestamp,filepath)
            
    def send(self,filepath):
        dst = self.gamedir.joinpath(filepath)
        src = gamedir.joinpath(filepath)
        dst.write_text(dst.text())
        name = makeName(filepath)
        self.update(name,int(time.time()),filepath)
        
        
### FILE DATABASE INTERFACE ###
class FileDB(object):
    def __init__(self,dbpath):
        self.dir = dbpath.joinpath('addons/eventscripts/globalconfigs/data')
        if not self.dir.isdir():
            self.dir.mkdir()
        self.gamedir = dbpath
        
    def add(self,filepath):
        filepath = filepath
        name = makeName(filepath)
        timestamp = 0
        f = self.dir.joinpath('%s.info' % name)
        f.write_text('%s %s' % (timestamp,filepath))
    
    def delete(self,filepath):
        filepath = filepath
        name = makeName(filepath)
        f = self.dir.joinpath('%s.info' % name)
        if f.isfile():
            f.remove()
    
    def getChanged(self):
        changed = []
        for name in self.dir.files('*.info'):
            timestamp,filepath = name.text().split(' ',1)
            timestamp = int(timestamp)
            fpath = self.gamedir.joinpath(filepath)
            if int(fpath.getmtime()) != timestamp:
                changed.append(filepath)
        return changed 
    
    def check(self,webanswer):
        timestamp,filepath = webanswer.split(' ',1)
        name = makeName(filepath)
        timestamp = int(timestamp)
        f = self.dir.joinpath('%s.info' % name)
        if not f.isfile():
            return False
        result = int(f.text().split(' ',1)[0])
        return result == timestamp # False= file will be updated!
    
    def getFilelist(self):
        result = []
        self.iUpdate()
        for f in self.dir.files('*.info'):
            timestamp,filepath = f.text().split(' ',1)
            result.append('%s %s' % (timestamp,filepath))
        return result
    
    def iUpdate(self):
        changed = self.getChanged()
        for filepath in changed:
            timestamp = int(self.gamedir.joinpath(filepath).getmtime())
            name = makeName(filepath)
            self.update(name,timestamp,filepath)
    
    def update(self,name,timestamp,filepath):
        f = self.dir.joinpath('%s.info' % name)
        f.write_text('%s %s' % (timestamp,filepath))
        
    def getFile(self,webanswer):
        timestamp,filepath = webanswer.split(' ',1)
        timestamp = int(timestamp)
        name = makeName(filepath)
        src = self.gamedir.joinpath(filepath)
        dst = gamedir.joinpath(filepath)
        dst.write_text(src.text())
        own.update(name,timestamp,filepath)
        
    def send(self,filepath):
        dst = self.gamedir.joinpath(filepath)
        src = gamedir.joinpath(filepath)
        dst.write_text(dst.text())
        name = makeName(filepath)
        self.update(name,int(time.time()),filepath)
    
### CONFIGURATIONS ###
class dyncfg(dict):
    def __init__(self,f,cvar_prefix='',default={}):
        self.__filepath__ = path.path(f)
        self.__filepath__.touch()
        self.__cvarprefix__ = cvar_prefix
        self.__d__ = {}
        self.__cvars__ = []
        self.sync()
        if default:
            for key in filter(lambda x: x not in self,default.keys()):
                self[key] = default[key]

    def __getitem__(self,s):
        s = str(s)
        if s in self.__d__:
            return self.__d__[s]
        L = self.__filepath__.lines(retain=False)
        for line in L:
            if not line.startswith('//'):
                var,val = line.split('=',1)
                var = var.strip()
                if val.count('//') == 1:
                    val = val.split('//')[0].strip()
                else:
                    val = val.strip()
                if var == s:
                    self.__d__[var] = val
                    self.__cvars__.append(self.__cvarprefix__ + var)
                    return val
        return None
    
    def __contains__(self,s):
        if s in self.__d__:
            return True
        L = self.__filepath__.lines(retain=False)
        for line in L:
            if line.startswith(s):
                return True
        return False

    def __setitem__(self,s,v):
        s = str(s)
        v = str(v)
        self.__d__[s] = v
        self.__cvars__.append(self.__cvarprefix__ + s)
        L = self.__filepath__.lines(retain=False)
        done = False
        for line in L:
            if not line.startswith('//') and line.count('=') != 0:
                if line.startswith(s):
                    if line.count('//') == 1:
                        L[L.index(line)] = '%s = %s //%s' % (s,v,line.split('//')[1])
                    else:
                        L[L.index(line)] = '%s = %s' % (s,v)
                    done = True
                    break
        if not done:
            L.append('%s = %s' % (s,v))
        self.__filepath__.write_lines(L)
        self.serverVar(s,v)
        
    def serverVar(self,s,v):
        es.ServerVar(self.__cvarprefix__ + s,v).addFlag('notify')

    def sync(self):
        self.__d__ = {}
        self.__cvars__ = []
        L = self.__filepath__.lines(retain=False)
        for line in L:
            if not line.startswith('//') and line.count('=') != 0:
                var,val = line.split('=')
                var = var.strip()
                if val.count('//') == 1:
                    val = val.split('//')[0].strip()
                else:
                    val = val.strip()
                self.__d__[var] = val
                self.serverVar(var,val)
                self.__cvars__.append(self.__cvarprefix__ + var)
        self.__filepath__.write_lines(L)
        
    def as_bool(self,s):
        val = self[s]
        if not val:
            return False
        else:
            if val.lower() in ('1','true','on'):
                return True
        return False
    
    def as_path(self,s):
        return path.path(self[s])            
                
    def keys(self):
        return self.__d__.keys()
        
    def cvars(self):
        return self.__cvars__
    
    def cvarchange(self,var,val):
        var = var[len(self.__cvarprefix__):]
        self[var] = val
   
def server_cvar(ev):
    if ev['cvarname'] in dcfg.cvars():
        dcfg.cvarchange(ev['cvarname'],ev['cvarvalue'])
    
default = {'sqlite':'1','local':'0','localpath':'','url':'','password':'','mapstart':'1'}
dcfg = dyncfg(gamedir.joinpath('cfg/globalconfigs.cfg'),'gc_',default)